/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  The official website and doumentation for xnec2c is available here:
 *    https://www.xnec2c.org/
 */

#ifndef SOMNEC_H
#define SOMNEC_H    1

#include "common.h"

#define PTP     .6283185308
#define PI10    31.41592654
#define TOSP    1.128379167
#define SP      1.772453851
#define ACCS    1.0E-12
#define GAMMA   .5772156649
#define C1      -.02457850915
#define C2      .3674669052
#define C3      .7978845608
#define P10     .0703125
#define P20     .1121520996
#define Q10     .125
#define Q20     .0732421875
#define P11     .1171875
#define P21     .1441955566
#define Q11     .375
#define Q21     .1025390625
#define POF     .7853981635
#define MAXH    20
#define CRIT    1.0E-4
#define NM      131072
#define NTS     4

#endif

